#include "GroupFromAnnotation.h"
#include "Burden.h"
#include "VT.h"
#include "QuickIndex.h"

String GroupFromAnnotation::groupFile = "";
String GroupFromAnnotation::vcfInput = "";
String GroupFromAnnotation::function = "";
String GroupFromAnnotation::freqFile = "";
bool GroupFromAnnotation::freqOut = false;


GroupFromAnnotation::GroupFromAnnotation()
{}

GroupFromAnnotation::~GroupFromAnnotation()
{
   //   if(chrList) delete [] chrList;
   //   if(posList) delete [] posList;
   //   if(alleleList) delete [] alleleList;
}

//freqFile has to be in the format of marker\tfreq
//marker has to be in the same format as in groupFile
void GroupFromAnnotation::GetFreqFromFile()
{
   StringArray tmp;
   FILE * file = fopen(freqFile,"r");
   String buffer;
   while (!feof(file))
   {
      buffer.ReadLine(file);
      tmp.Clear();
      tmp.AddTokens(buffer, "\t ");
      markerFreqHash.SetDouble(tmp[0],tmp[1].AsDouble());
   }
   fclose(file);
}

//TODO: make the grouping more robust to non-sorted variants
void GroupFromAnnotation::GetGroupFromFile()
{
   //Fill in annoGroups.
   StringArray tmp,tmp2;
   FILE * file = fopen(groupFile,"r");
   if(file==NULL)
   {
      error("Groupfile %s can not be opened.\n",groupFile.c_str());
   }
   String buffer;
   int line = 0;
   while (!feof(file))
   {
      buffer.ReadLine(file);
      tmp.Clear();
      tmp.AddTokens(buffer, SEPARATORS);
      annoGroups.Push(tmp[0]);
      line++;
   }
   fclose(file);
   geneCount = line;

   //Fill in SNPlist.
   chrList = new StringArray [line];
   posList = new IntArray [line];
   alleleList = new StringArray [line];
   uniqueChr =new StringArray [line];
   startPos =new IntArray [line];
   endPos =new IntArray [line];

   FILE * samefile = fopen(groupFile,"r");
   line = 0;
   StringArray chr,allele;
   IntArray pos;
   while (!feof(samefile))
   {
      chr.Clear();
      pos.Clear();
      allele.Clear();
      buffer.ReadLine(samefile);
      tmp.Clear();
      tmp.AddTokens(buffer, "\t ");
      chrList[line].Dimension(0);
      posList[line].Dimension(0);
      alleleList[line].Dimension(0);
      for(int i=1;i<tmp.Length();i++)
      {
	 tmp2.Clear();
	 if(freqOut)
	 {
	    double freq = markerFreqHash.Double(tmp[i]);
	    if(freq == _NAN_)
	    {
	       tmp2.AddTokens(tmp[i],"_");
	       StringArray alleles;
	       alleles.AddTokens(tmp2[1],"/");
	       String newMarker;
	       newMarker = tmp2[0] + "_" + alleles[1] + "/" + alleles[0]; 
	       freq = markerFreqHash.Double(newMarker);
	    }

	    if(freq != _NAN_ && freq>0.5)
	       freq = 1.0-freq;

	    if(freq!=_NAN_ && (freq<Burden::mafMin || freq>Burden::mafMax))
	       continue;
	    tmp2.Clear();
	 }
	 tmp2.AddTokens(tmp[i],":_/");
	 chr.Push(tmp2[0]);
	 pos.Push(tmp2[1].AsInteger());
	 allele.Push(tmp2[2]+"/"+tmp2[3]);
      }
      //SORT chr, pos, and allele and copy to chrList[line] and posList[line]
      QuickIndex chrIdx(chr);
      uniqueChr[line].Push(chr[chrIdx[0]]);
      IntArray start;//start and end of unique chr
      IntArray end;
      start.Push(0);

      for(int m=1;m<chr.Length();m++)
      {
	 if(chr[chrIdx[m]]!=uniqueChr[line][uniqueChr[line].Length()-1])
	 {
	    uniqueChr[line].Push(chr[chrIdx[m]]);
	    start.Push(m);
	    end.Push(m-1);
	 }
      }
      end.Push(chr.Length()-1);
/*
      printf("uniqueChr is:\n");
      for(int m=0;m<uniqueChr[line].Length();m++)
      {
	 printf("%s\t",uniqueChr[line][m].c_str());
      }
      printf("\n");
*/
      if(uniqueChr[line].Length()==1)
      {
	 //only have to sort pos
	 QuickIndex posIdx(pos);
	 startPos[line].Push(pos[posIdx[0]]);
	 endPos[line].Push(pos[posIdx[pos.Length()-1]]);
	 for(int m=0;m<pos.Length();m++)
	 {
	    posList[line].Push(pos[posIdx[m]]);
	    chrList[line].Push(chr[posIdx[m]]);
	    alleleList[line].Push(allele[posIdx[m]]);
	 }
      }
      else
      {
	 //sort within each chromosome
	 IntArray tmp;
	 StringArray tmpallele;
	 for(int l=0;l<uniqueChr[line].Length();l++)
	 {
	    tmp.Clear();
	    tmpallele.Clear();
	    if(start[l]==end[l])
	    {
	       chrList[line].Push(chr[chrIdx[start[l]]]);
	       posList[line].Push(pos[chrIdx[start[l]]]);
	       alleleList[line].Push(allele[chrIdx[start[l]]]);
	       startPos[line].Push(pos[chrIdx[start[l]]]);
	       endPos[line].Push(pos[chrIdx[start[l]]]);
	    }
	    else
	    {
	       for(int s=start[l];s<=end[l];s++)
	       {
		  tmp.Push(pos[chrIdx[s]]);
		  tmpallele.Push(allele[chrIdx[s]]);
	       }
	       QuickIndex subPosIdx(tmp);
	       startPos[line].Push(tmp[subPosIdx[0]]);
	       endPos[line].Push(tmp[subPosIdx[tmp.Length()-1]]);
	       for(int s=0;s<tmp.Length();s++)
	       {
		  chrList[line].Push(uniqueChr[line][l]);
		  posList[line].Push(tmp[subPosIdx[s]]);
		  alleleList[line].Push(tmpallele[subPosIdx[s]]);
	       }
	    }
	 }
      }
/*
printf("start position is:\n");
      for(int m=0;m<start.Length();m++)
      {
         printf("%d\t",startPos[line][m]);
      }
      printf("\n");
printf("end position is:\n");
      for(int m=0;m<start.Length();m++)
      {
         printf("%d\t",endPos[line][m]);
      }
      printf("\n");

      printf("chrList[line] is:\n");
      for(int m=0;m<chrList[line].Length();m++)
      {
	 printf("%s\t",chrList[line][m].c_str());
      }
      printf("\n");

      printf("posList[line] is:\n");
      for(int m=0;m<posList[line].Length();m++)
      {
	 printf("%d\t",posList[line][m]);
      }
      printf("\n");
      printf("alleleList[line] is:\n");
      for(int m=0;m<alleleList[line].Length();m++)
      {
	 printf("%s\t",alleleList[line][m].c_str());
      }
      printf("\n");
*/
      line++;
   }
   fclose(samefile);
}

void GroupFromAnnotation::GetGroupFromSardiNiaFormat()
{ 
   //fill in geneHash with start and end positions
   reader.open(vcfInput,header);
   reader.readVcfIndex();
   geneChrom.Clear();
   std::string gene;
   const std::string* txGNstr;
   while(reader.readRecord(record))
   {
      VcfRecordInfo & info = record.getInfo();
      txGNstr =  info.getString("txGN");
      if(txGNstr == NULL)
      {
	 continue;
      }
      //parse the txGNstr
      size_t found = txGNstr->find('/');
      size_t oldpos = found;
      if(found == std::string::npos)
      {
	 gene = *txGNstr;
	 HashInGenes(gene);
      } else {
	 gene = txGNstr->substr(0,found);
	 HashInGenes(gene);
	 oldpos=found;
	 found = txGNstr->find('/',found+1);
	 while(found != std::string::npos)
	 {
	    gene = txGNstr->substr(oldpos+1,found-oldpos-1);
	    HashInGenes(gene);
	    oldpos=found;
	    found = txGNstr->find('/',found+1);
	 }
	 gene = txGNstr->substr(oldpos+1,txGNstr->length()-oldpos-1);
	 HashInGenes(gene);
      }
   }

   chrList = new StringArray [geneCount];
   posList = new IntArray [geneCount];
   alleleList = new StringArray [geneCount];

   //fill in chrList posList alleleList annoGroups

   const std::string* FUNCstr;

   for(int i=0;i<geneHash.Capacity();i++)
   {
      if(!geneHash.SlotInUse(i)) 
      {
	 continue;
      }

      String gene = geneHash[i];
      annoGroups.Push(geneHash[i]);
      int value = geneHash.Integer(geneHash[i]);
      chrList[value].Dimension(0);
      posList[value].Dimension(0);
      alleleList[value].Dimension(0);
      //printf("grouping gene %s, %d, start=%d, end=%d\n",gene.c_str(),value,start[value],end[value]);

      reader.set1BasedReadSection(geneChrom[value].c_str(),start[value],end[value]+1);

      while(reader.readRecord(record))
      {
	 VcfRecordInfo & info = record.getInfo();
	 String ref = record.getRefStr();
	 String alt = record.getAltStr();
	 String allele = ref + "/" + alt;
	 String chr_current = record.getChromStr();
	 int pos_current = record.get1BasedPosition();

	 if(freqOut)
	 {
	    String markerName;
	    markerName = chr_current + ":" + pos_current + "_" + allele;
	    double freq = markerFreqHash.Double(markerName);
	    if(freq==_NAN_)
	    {
	       markerName = chr_current + ":" + pos_current + "_" + alt + "/" + ref;
	       freq = markerFreqHash.Double(markerName);
	    }
	    if(freq != _NAN_ && freq>0.5)
	       freq = 1.0-freq;
	    if(freq != _NAN_ && (freq < Burden::mafMin || freq > Burden::mafMax))
	    {
	       continue;
	    }
	 }

	 txGNstr =  info.getString("txGN");
	 FUNCstr =  info.getString("FUNC");
	 if(txGNstr == NULL)
	 {
	    continue;
	 }
	 bool found_gene=false;
	 size_t found = txGNstr->find(gene.c_str());
	 if(found == std::string::npos)
	 {
	    continue;
	 }

	 if(txGNstr->length()==gene.Length())
	 {
	    found_gene=true;
	 }
	 if(txGNstr->length()>gene.Length())
	 {
	    StringArray tmp;
	    tmp.AddTokens((*txGNstr).c_str(),SEPARATORS);
	    for(int i=0;i<tmp.Length();i++)
	    {
	       if(tmp[i].Length()!=gene.Length())
		  continue;
	       else
	       {
		  if(tmp[i].FastCompare(gene)==0)
		  {
		     found_gene=true;
		     break;
		  }
		  else
		  {
		     continue;
		  }
	       }
	    }
	    if(!found_gene) continue;
	 }
	 if(function!="")
	 {
	    if(FUNCstr==NULL) continue;

	    StringArray func;
	    func.AddTokens(function,"/");
	    bool foundFunc=false;
	    for(int i=0;i<func.Length();i++)
	    {
	       if(func[i]=="syn")
	       {
		  if(FUNCstr->find(func[i].c_str())==std::string::npos)
		     continue;
		  else if(FUNCstr->find("nonsyn")!=std::string::npos)
		     continue;
		  else
		  {
		     foundFunc=true;
		     break;
		  }
	       }
	       else
	       {
		  if(FUNCstr->find(func[i].c_str())==std::string::npos)
		     continue;
		  else
		  {
		     foundFunc=true;
		     break;
		  }
	       }
	    }
	    if(!foundFunc)
	       continue;
	 }

	 chrList[value].Push(chr_current);
	 posList[value].Push(pos_current);
	 alleleList[value].Push(allele);
      }
   }
   reader.close();
}

void GroupFromAnnotation::HashInGenes(std::string & gene)
{
   int geneidx = geneHash.Integer(gene.c_str());
   if(geneidx == -1) {
      geneHash.SetInteger(gene.c_str(),geneCount);
      start.resize(geneCount+1);
      end.resize(geneCount+1);
      start[geneCount] = record.get1BasedPosition();
      end[geneCount] = start[geneCount];
      geneChrom.Push(record.getChromStr());
      geneCount++;
   } 
   else 
   {
      end[geneidx] = record.get1BasedPosition();
   }
}

void GroupFromAnnotation::GetAllGenenamesVCF()
{
   reader.open(vcfInput,header);

   geneCount =0;
   geneChrom.Clear();
   std::string gene;
   const std::string* txGNstr;
   while(reader.readRecord(record))
   {
      VcfRecordInfo & info = record.getInfo();
      txGNstr =  info.getString("txGN");
      if(txGNstr == NULL)
      {
	 continue;
      }
      //parse the txGNstr
      size_t found = txGNstr->find('/');
      size_t oldpos = found;
      if(found == std::string::npos)
      {
	 gene = *txGNstr;
	 HashInGenes(gene);
      } else {
	 gene = txGNstr->substr(0,found);
	 HashInGenes(gene);
	 oldpos=found;
	 found = txGNstr->find('/',found+1);
	 while(found != std::string::npos)
	 {
	    gene = txGNstr->substr(oldpos+1,found-oldpos-1);
	    HashInGenes(gene);
	    oldpos=found;
	    found = txGNstr->find('/',found+1);
	 }
	 gene = txGNstr->substr(oldpos+1,txGNstr->length()-oldpos-1);
	 HashInGenes(gene);
      }
   }
   reader.close();
}

void GroupFromAnnotation::Run()
{ 
   geneCount = 0;
   if(Burden::CMC_counts || Burden::SKAT_BETA || Burden::SKAT_UW ||Burden::SKAT_MB|| VT::asymptotic || VT::permuteBool)
   {
      if(freqFile != "")
      {
	 freqOut=true;
	 GetFreqFromFile();
      }
      if(groupFile =="")
	 GetGroupFromSardiNiaFormat();
      else
	 GetGroupFromFile();
   }
}


